;; Generated package description from bind-key.el  -*- no-byte-compile: t -*-
(define-package "bind-key" "2.4.1.0.20240210.91008" "A simple way to manage personal keybindings" '((emacs "24.3")) :commit "7f3baf352bad03de50135556a561af0c7fb1bd6a" :authors '(("John Wiegley" . "johnw@newartisans.com")) :maintainer '("John Wiegley" . "johnw@newartisans.com") :keywords '("keys" "keybinding" "config" "dotemacs" "extensions") :url "https://github.com/jwiegley/use-package")
